"""
This module contains the simulation engine class.
"""


# %%
# Import required packages
import numpy as np
from .context import Context
from .schedule import Schedule
from numpy.typing import NDArray


# %%
class Engine:
    """
    A class for managing the simulation engine. The engine is the main executor for the simulation and container for the
    context. The simulation is executed using the *run()* method.

    Parameters
    ----------
    modelTime : float
        The total time for which the simulation will run.
    dimensions : NDArray[np.float64]
        The dimensions of the simulation environment, based on 3D representation. The order of XYZ boundaries is done
        the following: *[X-min, Y-min, Z-min, X-max, Y-max, Z-max]*
    context : Context, optional
        The context to be used, if custom. Default is to use the built-in context.

    Attributes
    ----------
    tick : float
        The current simulation tick.
    modelTime : float
        The total time for which the simulation will run.
    dimensions : NDArray[np.float64]
        The dimensions of the simulation environment.
    context : Context | Any
        The context containing the agents and environment of the simulation. Can also be a child class of Context class.
    schedule : Schedule
        The schedule managing the actions to be executed.
    """

    def __init__(
        self, modelTime: float, dimensions: NDArray[np.float64], context: Context = None
    ):
        self.tick = 0.0
        self.modelTime = modelTime
        self.schedule = Schedule(self.tick)
        self.dimensions = dimensions

        # Initialize context, allowing for custom context to be passed
        if context is None:
            self.context = Context(dimensions)
        else:
            self.context = context

    def run(self):
        """
        Runs the simulation until reached model time or schedule is empty
        """
        while self.tick <= self.modelTime and self.schedule.get_size() > 0:
            self.tick = self.schedule.step()
            # print(self.tick)

        # print("RUN COMPLETED!")

    def abort(self):
        """
        Aborts the simulation and prints the stopped time.
        """
        self.schedule.clear_schedule()
        print(f"Stopped at: {self.tick}")

    def get_context(self) -> Context:
        """
        Get the context connected to the simulation.

        Returns
        -------
        context : Context
            The simulation context.
        """
        return self.context



"""
This module contains the simulation context class.
"""

# %%
# Import required packages
import numpy as np
from numpy.typing import NDArray

from .agent import Agent


# %%
class Context:
    """
    A class representing the simulation context, managing agents within a defined space.

    Parameters
    ----------
    dimensions: NDArray[np.float64]
        The dimensions of the simulation environment.

    Attributes
    ----------
    dimensions: NDArray[np.float64]
        The dimensions of the simulation environment.
    agents: list
        A list of agents present in the simulation.

    """

    # Initializes the context with dimensions and empty agent list
    def __init__(self, dimensions: NDArray[np.float64]):
        self._positions_cache = None
        self.dimensions = dimensions
        self.agents = list()
        self._dirty = True

    def add_agent(self, agent: Agent):
        """
        Adds an agent to the context.

        Parameters
        ----------
        agent : Agent
            The agent to add.
        """
        self.agents.append(agent)
        self._dirty = True

    def remove_agent(self, agent: Agent):
        """
        Removes a specified agent from the context.

        Parameters
        ----------
        agent : Agent
            The agent to be removed.
        """
        # Finds the right agent in the list and removes it
        self.agents.remove(agent)
        self._dirty = True

    def mark_dirty(self):
        self._dirty = True

    # Checks if an object is of a specific class name
    @staticmethod
    def check_class(obj, name) -> bool:
        if obj.__class__.__name__ == name:
            return True
        else:
            return False

    # Getters
    def get_agents(self) -> list:
        """
        Gets all agents in the context.

        Returns
        -------
        agents : list
            A list of agents.
        """
        return self.agents

    def get_agents_by_class(self, class_name: str) -> list:
        """
        Gets all agents of a specific class.

        Parameters
        ----------
        class_name : str
            The name of the class

        Returns
        -------
        agents : list
            A list of agents.
        """
        return [agent for agent in self.agents if self.check_class(agent, class_name)]

    def get_dimensions(self) -> NDArray[np.float64]:
        """
        Gets the dimensions of the context.

        Returns
        -------
        dimensions : NDAArray[np.float64]
            The dimensions.
        """
        return self.dimensions

    def get_positions_array(self):
        """
        Returns a (n, dim) numpy array of agent positions.
        Cached until an agent calls context.mark_dirty().

        Returns
        -------
        pos_array : np.ndarray
            The position array.
        """
        if self._dirty:
            if not self.agents:
                self._positions_cache = np.empty((0, self.dimensions.size // 2))
            else:
                # list comprehension into vstack once per rebuild
                self._positions_cache = np.vstack(
                    [a.get_position() for a in self.agents]
                )
            self._dirty = False
        return self._positions_cache

    def get_agent_count(self, classes: list = None) -> dict:
        """
        Gets the agent count for each agent type based on class.

        Parameters
        ----------
        classes : list, optional
            A List of classes to count. If *None*, then alla classes are counted.

        Returns
        -------
        counts : dict
            A dictionary with each agent class and their count.
        """
        entry = dict()

        # if no classes provided, return total count for each agent type
        if not classes:
            unique_classes = set(obj.__class__.__name__ for obj in self.agents)
            for cls in unique_classes:
                a = sum(self.check_class(obj, cls) for obj in self.agents)
                entry[cls] = a
            return entry

        for arg in classes:
            # print(arg)
            a = sum(self.check_class(obj, arg) for obj in self.agents)
            # print(a)
            entry[arg] = a

        # print(entry)
        return entry
